package org.speleologist.common;

/*
 * Surveyor - the Poor Man's Digital Survey Device
 * Copyright Chris Hunter 2012 
 * email chris_at_arnside[dot]net
 * 
 * This file is part of Surveyor.

    Surveyor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Surveyor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Surveyor.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.BufferedReader;
import android.graphics.Point;
import android.graphics.Rect;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.Stack;


public class LcdDigitOcr {

	private int[][] pixels;
	private int imgWidth;
	private int imgHeight;

	private int[][] segBounds; // min and max X/Y for the digit area
	private LineTracer lt;
	private float threshold;
	

	// we order the segments from top left going down then across
	// this represents which segments should be on for a given digit
	private static BitSet[] d = new BitSet[10];
	static {
		for (int i = 0; i < 10; i++)
			d[i] = new BitSet();
		d[0].set(0);d[0].set(1);d[0].set(2);d[0].set(4);d[0].set(5);d[0].set(6);
		d[1].set(5);d[1].set(6);d[2].set(1);d[2].set(2);d[2].set(3);
		d[2].set(4);d[2].set(5);
		d[3].set(2);d[3].set(3);d[3].set(4);d[3].set(5);d[3].set(6);
		d[4].set(0);d[4].set(3);d[4].set(5);d[4].set(6);d[5].set(0);
		d[5].set(2);d[5].set(3);d[5].set(4);d[5].set(6);
		d[6].set(0);d[6].set(1);d[6].set(2);d[6].set(3);d[6].set(4);d[6].set(6);
		d[7].set(2);d[7].set(5);d[7].set(6);
		d[8].set(0);d[8].set(1);d[8].set(2);d[8].set(3);d[8].set(4);d[8].set(5);d[8].set(6);
		d[9].set(0);d[9].set(2);d[9].set(3);d[9].set(4);d[9].set(5);d[9].set(6);
	}

		
	public LcdDigitOcr(int[][] yuvPixels, int[][] digitArea, int threshold) {
		this(digitArea, threshold);
		
		this.imgWidth =  yuvPixels.length;
		this.imgHeight = yuvPixels[0].length;
		
		brightnessToMonochrome(yuvPixels);
		//this.traceSegments();

	}
	
	
	public LcdDigitOcr(int[][] digitArea, int threshold) {
		segBounds = digitArea;
		
		this.threshold = (float)threshold/100f;
		//this.traceSegments();

	}

	
	/**
	 * read the value from the given brightness data
	 * 
	 * @param lum
	 * @param saveImage
	 * @return
	 */
	public String readValue(int[][] lum) {		
		this.brightnessToMonochrome(lum);
		traceSegments();
		
		
		return readDigits();
	}
		
	
	/**
	 * Use the line tracer to trace the segments
	 * 
	 * @return returns the list of path groups for display and debug if required
	 */
	public ArrayList<ArrayList<int[]>> traceSegments() {
		
		lt = new LineTracer(pixels);
		lt.trace();
		return lt.pathGroups;
	}
	

	public String readDigits() {
		ArrayList<Bounds> bounds = new ArrayList<Bounds>();
		//sort out the resulting polygons
		for(ArrayList<int[]> paths : lt.pathGroups){
			Bounds b = new Bounds(paths);
			bounds.add(b);			
		}
		
		Collections.sort(bounds);
		
		//remove any entirely inside another, concatenate adjacent and over laps
		ArrayList<Bounds> forRemoval = new ArrayList<Bounds>();
		for(Bounds b: bounds) {			
			if(forRemoval.contains(b)) continue;
			
			dumpBounds(b);
			
			//remove any speckles and the decimal point
			if(Math.abs(b.xmin - b.xmax)<10
			&& Math.abs(b.ymin - b.ymax)<10) {
				if(!forRemoval.contains(b)) forRemoval.add(b);
			}
			else {
				for(int j=bounds.indexOf(b)+1; j<bounds.size(); j++) {					
					if(forRemoval.contains(bounds.get(j))) continue;
					
					//dumpBounds(bounds.get(j));
					
					if(Math.abs(bounds.get(j).xmin - bounds.get(j).xmax)<10
					&& Math.abs(bounds.get(j).ymin -bounds.get(j).ymax)<10) {
						if(!forRemoval.contains(bounds.get(j))) forRemoval.add(bounds.get(j));
					}
					else {
						if(bounds.get(j).isInside(b)) {
							b.concatenate(bounds.get(j));
							if(!forRemoval.contains(bounds.get(j))) forRemoval.add(bounds.get(j));
						}				
						else if(bounds.get(j).isOverlapping(b)
								|| bounds.get(j).isAdjacent(b)){
							b.concatenate(bounds.get(j));
							if(!forRemoval.contains(bounds.get(j))) forRemoval.add(bounds.get(j));
						}
					}//else not too small
				}//for j
				//System.out.println("New Bounds >>>>");
				dumpBounds(b);
			}
		}
		//System.out.println("\n");
		
		for(Bounds b : forRemoval) {
			bounds.remove(b);
		}
		
		Collections.sort(bounds);
		
		//get the width of a digit - take top right of first digit
		//to top right of last (can't use left if first char is a 1...)
		for(Bounds b : bounds) {
			b.findSegMinMax();
		}
		int numDig = bounds.size();
		if (numDig<1) return "____";
		
		int diff =  bounds.get(numDig-1).topR.x - bounds.get(0).topR.x;
		//assume the digit is 80% of the space
		int digWidth = (int) (diff * 0.8/(numDig-1));		
		
		//resolve it to digits
		String digits = "";
		for(Bounds b : bounds) {
			
			//dumpBounds(b);
			
			//is it narrow
			if(b.xmax-b.xmin < digWidth-4) {
				if(Math.abs((b.topL.x-b.topR.x)-(b.botL.x-b.botR.x) ) <3)
					digits += "1";
				else
					digits += "7";
			}
			else {
				//find the centre				
				int qtr = (b.ymax-b.ymin)/4;
				int cx = b.xmin + (b.xmax - b.xmin)/2;				
				int cy= b.ymin + (2*qtr);
								
				BitSet segs = new BitSet();
				
				//top left - go -ve Y a bit and then check -ve X
				if(checkX(b.topL.x, cx-3, cy-qtr)) segs.set(0);
				//bot left - go +ve Y a bit and then check -ve X
				if(checkX(b.botL.x, cx-3, cy+qtr)) segs.set(1);
				//top middle scan from edge in along Y
				if(checkY(cx, b.topL.y, cy-4)) segs.set(2);
				//is the centre on?
				if(checkY(cx, cy-2, cy+2)) segs.set(3);
				//bot middle scan from beyond centre to edge in along X
				if(checkY(cx, cy+qtr, b.botL.y)) segs.set(4);
				//top right
				if(checkX(cx+3, b.topR.x, cy-qtr)) segs.set(5);
				//bot right - go +ve Y a bit and then check +ve Y
				if(checkX(cx+3, b.botR.x, cy+qtr)) segs.set(6);
				
				int i;
				for (i = 0; i < 10; i++) {
					if (d[i].equals(segs)) {
						digits += String.valueOf(i);
						break;
					}
				}
				if (i == 10)
					digits += "_";
			}
		}
		
		//System.out.println("\n>>>>> " + digits + " <<<<<");
		
		return digits;
	}
	
	
	private boolean checkX(int start, int limit, int cy) {
		boolean result = false;		
		
		if(cy>=0 && cy<pixels[0].length) {
			for(int n=start; n<=limit; n++) {
				if(n>=0 && n<pixels.length) {
					if(pixels[n][cy]==0) {
						result = true;
						break;
					}
				}
			}
		}
		return result;
	}
	
	private boolean checkY(int cx, int start, int limit) {
		boolean result = false;
		if(cx>=0 && cx<pixels.length) {
			for(int n=start; n<=limit; n++) {
				if(n>=0 && n<pixels[0].length) {
					if(pixels[cx][n]==0) {
						result = true;
						break;
					}
				}
			}
		}
		return result;
	}
	
	
	public int[][] getPixels() {
		return pixels;
	}
	
	
	private void dumpBounds(Bounds b) {
		System.out.println("\n");
		System.out.format("\t%d\t%d\t%d\t%d\t%n", 
				b.xmin, b.ymin,	b.xmax, b.ymax);
		dumpPixels(b.xmin, b.ymin, b.xmax, b.ymax);
		System.out.println("\n");
	}
	
	private void dumpPixels(int xmin, int ymin, int xmax, int ymax) {
		
		for(int y=ymin; y<ymax; y++) {
			for(int x=xmin; x<xmax; x++) {			
				System.out.print(pixels[x][y]);
			}
			System.out.println();
		}
	}



	/**
	 * process ints representing brightness into just black or white pixels
	 * for the area containing the digits
	 * 
	 * @param rgb
	 */
	public int[][] brightnessToMonochrome(int[][] lum) {

		int xs = 0, ys = 0;
		int xf = imgWidth, yf = imgHeight;
		if (segBounds != null) {
			xs = segBounds[0][0];
			xf = segBounds[1][0];
			ys = segBounds[0][1];
			yf = segBounds[1][1];
		}

		// trim it down to just the digits
		pixels = new int[xf - xs][yf - ys];
		for (int y = ys; y < yf; y++) {
			for (int x = xs; x < xf; x++) {
				pixels[x - xs][y - ys] = lum[x][y];
			}
		}
		
		//dumpPixels(0,0, pixels.length, pixels[0].length);		
		if(pixels[0].length > pixels.length) {
			int [][] tmp = new int[yf - ys][xf - xs];
			//it's rotated x should be > y
			for(int larger=0; larger<pixels[0].length; larger++) {
				for(int smaller=0; smaller<pixels.length; smaller++) {
					tmp[pixels[0].length-1-larger][smaller] = pixels[smaller][larger];
				}
			}		
			pixels = tmp;
		}		
		//dumpPixels(0,0, pixels.length, pixels[0].length);
		
		int yo = pixels[0].length/2;
		int sum = 0;
		for(int x= 0; x<pixels.length; x++ ) {
			sum += pixels[x][yo];
		}
		int aveB = sum/pixels.length;

		int[][] thold = new int[pixels.length][pixels[0].length];
		for (int y = 0; y < pixels[0].length; y++) {
			for (int x = 0; x < pixels.length; x++) {
				// System.out.format("%d\t%d%n", aveB, pixels[x][y]);
				if (pixels[x][y] < aveB * threshold)
					thold[x][y] = 0;
				else
					thold[x][y] = 1;
			}
		}

		pixels = thold;
		
		return pixels;
	}
	
	
	private class Bounds implements Comparable<Bounds>{
		Point topL; //of the digit not of the coordinate space
		Point topR;
		Point botL;
		Point botR;
		
		int xmin = 999999999;
		int ymin = 999999999;
		int xmax = -999999999;
		int ymax = -999999999;
		
		ArrayList<int[]> lines;
		
		
		Bounds(ArrayList<int[]> lines) {
			topL = new Point(999999, -999999);
			topR = new Point(999999, 999999);
			botL = new Point(-999999, -999999);
			botR = new Point(-999999, -999999);
			
			this.lines = lines;
			
			findBounds(lines);
		}
		
		/**
		 * Find the extents of this digit
		 * 
		 * the digits are italic so we need to determine all 4 coners
		 * the digits are in there side
		 * @param point
		 */
		void findBounds(ArrayList<int[]> lines) {
			//cycle through and find the max/min
			for(int[] line : lines) {
				checkPoint(line);
			}			
		}
		
		
		/**
		 * find the top and bottom segment positions
		 */
		void findSegMinMax() {
			int txmin=999999;
			int txmax=-999999;
			int bxmin=999999;
			int bxmax=-999999;
			
			for(int[] line : lines) {
				if(Math.abs(line[1]-ymin)<10) {
					if(line[0]<txmin) txmin = line[0];
					if(line[0]>txmax) txmax = line[0];
					if(line[2]<txmin) txmin = line[2];
					if(line[2]>txmax) txmax = line[2];
				}
				
				if(Math.abs(line[1]-ymax)<10) {
					if(line[0]<bxmin) bxmin = line[0];
					if(line[0]>bxmax) bxmax = line[0];
					if(line[2]<bxmin) bxmin = line[2];
					if(line[2]>bxmax) bxmax = line[2];
				}
			}
			
			topL = new Point(txmin, ymin);
			topR = new Point(txmax, ymin);
			botL = new Point(bxmin, ymax);
			botR = new Point(bxmax, ymax);
		}
		
		
		void checkPoint(int[] point) {
			if(point[0]<xmin) xmin=point[0];
			if(point[1]<ymin) ymin=point[1];
			if(point[0]>xmax) xmax=point[0];
			if(point[1]>ymax) ymax=point[1];
			
			if(point[2]<xmin) xmin=point[2];
			if(point[3]<ymin) ymin=point[3];
			if(point[2]>xmax) xmax=point[2];
			if(point[3]>ymax) ymax=point[3];
		}


		
		public int compareTo(Bounds that) {
			
			//Most Significant Digit is at the left so lower x is higher in sort order
			if(this.xmax < that.xmax) return -1;
			else if(this.xmax > that.xmax) return 1;
			else return 0;
	
		}
		
		
		public boolean isInside(Bounds that) {
			boolean result = false;
			result =  (that.xmin>=this.xmin
					&& that.ymin>=this.ymin
					&& that.xmax<=this.xmax
					&& that.ymax<=this.ymax);
			
			if(!result) {
				result =  (this.xmin>=that.xmin
						&& this.ymin>=that.ymin
						&& this.xmax<=that.xmax
						&& this.ymax<=that.ymax);
			}
			return result;
		}
		
		public boolean isOverlapping(Bounds that) {
			Rect rthis = new Rect(xmin, ymin, xmax-xmin, ymax-ymin);
			Rect rthat = new Rect(that.xmin, that.ymin, that.xmax-that.xmin, that.ymax-that.ymin);
			
			return rthis.intersect(rthat);
		}
		
		public boolean isAdjacent(Bounds that) {
			boolean result = false;
			
			int dthis=0;
			int dthat=99999;
			int rthis = 0;
			int rthat = 999999;
			
			//check if the x's are aligned
			dthis = this.xmax-this.xmin;
			rthis = this.xmin;
			dthat = that.xmax-that.xmin;
			rthat = that.xmin;
			
			if(dthis>dthat) {
				result = (rthat>rthis-3) && (rthat< rthis+dthis+3);
			}
			else {
				result = (rthis>rthat-3) && (rthis< rthat+dthat+3);
			}
			
			return result;
		}
		
		public void concatenate(Bounds that) {
			if(that.xmin<xmin) xmin=that.xmin;
			if(that.ymin<ymin) ymin=that.ymin;
			if(that.xmax>xmax) xmax=that.xmax;
			if(that.ymax>ymax) ymax=that.ymax;
			
			this.lines.addAll(that.lines);
		}
	}
	

}
