package org.speleologist.common;

/*
 * Surveyor - the Poor Man's Digital Survey Device
 * Copyright Chris Hunter 2012 
 * email chris_at_arnside[dot]net
 * 
 * This file is part of Surveyor.

    Surveyor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Surveyor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Surveyor.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.Stack;

import android.content.res.Configuration;
import android.hardware.Sensor;
import android.hardware.SensorEvent;
import android.hardware.SensorEventListener;
import android.hardware.SensorManager;
import android.widget.TextView;

public class Orientation implements SensorEventListener {
    private float[] rotationMatrix = new float[16];   
    private Damper [] accels = { new Damper(), new Damper(), new Damper() };
    private Damper [] mags = { new Damper(), new Damper(), new Damper() };
    
    /* fix random noise by averaging tilt values */
    private float mLastPitch = 0.f;
    private float mLastYaw = 0.f;
    private float mLastRoll = 0.f;
    private OrientationListener listener;
        
    private boolean swapPitchRollLandscape;
    
    public Orientation() {
    	swapPitchRollLandscape=false;
    }
    
    
    public Orientation(OrientationListener listener) {
    	this();
    	this.listener = listener;
    }
    
    
    public void onSensorChanged(SensorEvent event) {
		boolean changed = false;
    	if (event.sensor.getType() == Sensor.TYPE_ACCELEROMETER) {
			updateAccel(event);
			changed = true;
		}
		if (event.sensor.getType() == Sensor.TYPE_MAGNETIC_FIELD) {
			updateMag(event);
			changed = true;
		}
		
		if(changed && listener!=null) listener.orientationChanged();
		
	}

	
	public void onAccuracyChanged(Sensor sensor, int accuracy) {
	}


    public void updateAccel(SensorEvent event) {
        damp(accels, event.values);
    }
    
    public void updateAccel(float[] values) {
    	damp(accels, values);
    }

    public void updateMag(SensorEvent event) {
    	damp(mags, event.values);
    }
    
    // testing
    public void updateMag(float[] values) {
    	damp(mags, values);
    }
    
    private void damp(Damper[] damper, float[] vals) {
    	for(int i=0; i<3; i++) {
    		damper[i].append(vals[i]);
    	}
    }
    
    
    public void computeOrientation(int screenOrient) {
    	
    	float[] aveAccels = new float[3];
    	for(int i=0; i<3; i++) aveAccels[i] = accels[i].avg();
    	
    	float[] aveMags = new float[3];
    	for(int i=0; i<3; i++) aveMags[i] = mags[i].avg();    	
    	
        if (SensorManager.getRotationMatrix(rotationMatrix, null, aveAccels, aveMags)) {
        	        	
        	//remap so that the yaw is correct both landscape and portrait
        	float[] remapMatrix = new float[16];
        	if(aveAccels[0]>5) {
        		//x is pointing down
        		SensorManager.remapCoordinateSystem(rotationMatrix,
                        SensorManager.AXIS_Z, SensorManager.AXIS_Y,
                        remapMatrix);
        	}
        	else if(aveAccels[0]<-5) {
        		//x is pointing up
        		SensorManager.remapCoordinateSystem(rotationMatrix,
                        SensorManager.AXIS_MINUS_Z, SensorManager.AXIS_Y,
                        remapMatrix);
        	}
        	else if(aveAccels[1]>5){
    			//y is pointing down
    			SensorManager.remapCoordinateSystem(rotationMatrix,
                        SensorManager.AXIS_X, SensorManager.AXIS_Z,
                        remapMatrix);          	
        	}
        	else if(aveAccels[1]<-5){
    			//y is pointing up
    			SensorManager.remapCoordinateSystem(rotationMatrix,
                        SensorManager.AXIS_MINUS_X, SensorManager.AXIS_MINUS_Z,
                        remapMatrix);      	
        	}
        	else if(aveAccels[2]<-5){
    			//Z is pointing up
    			SensorManager.remapCoordinateSystem(rotationMatrix,
                        SensorManager.AXIS_MINUS_X, SensorManager.AXIS_Y,
                        remapMatrix);      	
        	}
        	else {
        		//z is down
        		System.arraycopy(rotationMatrix, 0, remapMatrix, 0, 16);
        	}
        	
                  	        	
            float[] orientation = new float[3];
            SensorManager.getOrientation(remapMatrix, orientation);        	
            float yaw = (float) Math.toDegrees(orientation[0]);
            yaw = (yaw >= 0 )? yaw: yaw + 360;
            
            //use the non-remapped data for roll and pitch
            SensorManager.getOrientation(rotationMatrix, orientation);
            float tp = orientation[1];
            float tr = orientation[2];
		
            if(swapPitchRollLandscape && screenOrient == Configuration.ORIENTATION_LANDSCAPE){
            	tp = orientation[2];
            	tr = orientation[1];
            }
            
            float pitch = (float)Math.toDegrees(tp);
            //pitch = (float) ((pitch + 360.0) % 360.0);
            /*if (pitch <= -90) { 
            	pitch += (-2*(90+pitch)); 
            } 
            else if(pitch >= 90){ 
            	pitch += (2*(90 - pitch)); 
            }*/
            
            float roll = (float)Math.toDegrees(tr);
            //roll = (float) ((roll + 360.0) % 360.0);
            mLastYaw = yaw;
            mLastPitch = pitch;
            mLastRoll = roll;
                        
        }        
    }
    

    private class Damper {
        static final int AVERAGE_BUFFER = 20;
        private Stack<Float> vals = new Stack<Float>();

        public void append(float val) {
        	vals.push(val);
            if (vals.size() > AVERAGE_BUFFER)  
            	vals.remove(vals.firstElement());
            
            //return avg();
        }
        
        /** compute the average 
         * @return
         */        
        public float avg() {
            float sum = 0;
            for (float x: vals){
                sum += x;
            }
            float ave1 =  sum / vals.size();
            
            /*
            //std dev
    		sum = 0;
    		for (float x: vals) {
    			float dev = x - ave1;
                sum += dev*dev;
    		}
    		float sd = (float) Math.sqrt(sum/vals.size());
               		
            int aveCt = 0;
            sum = 0;
            for(float x: vals) {
            	if(Math.abs(ave1 - x) - sd < 2) {
            		aveCt++;
            		sum += x;
            	}
            }
            
            return sum/aveCt;
            */
            
            return ave1;
        }
        
        
    }


	public float getPitch() {
		return mLastPitch;
	}

	public float getYaw() {
		return mLastYaw;
	}

	public float getRoll() {
		return mLastRoll;
	}


	public float[] getAveMagFields() {
		float[] aveMags = new float[3];
    	for(int i=0; i<3; i++) aveMags[i] = mags[i].avg();
    	return aveMags;
	}
	
	
	public float[] getAveGravFields() {
		float[] aveGrav = new float[3];
    	for(int i=0; i<3; i++) aveGrav[i] = accels[i].avg();
    	return aveGrav;
	}


	public boolean isSwapPitchRollLandscape() {
		return swapPitchRollLandscape;
	}


	public void setSwapPitchRollLandscape(boolean swapPitchRollLandscape) {
		this.swapPitchRollLandscape = swapPitchRollLandscape;
	}
}
